package Our_Fireworks
{

import flash.display.DisplayObjectContainer;
import flash.display.Graphics;

/**
 * Author: H. James de St. Germain Date: Fall 2007
 * 
 * Updated and Ported to Java: Spring 2010
 * 
 * The Firework class ISA particle (which represents the physical location of the firework and how it moves). It also
 * contains a link to the "containing" GUI, where the firework draws itself. Further, it contains more information about
 * what it means to be a firework. This includes:
 * 
 * How long till it explodes, How many "sub-rockets" it has (payload) What color it is
 * 
 * We also keep a reference to the simulator so that we can add our children (payload).
 * 
 * NOTE: The Firework is a MOVABLE (based on the fact that a Particle is a MOVABLE)
 * 
 * 
 */

public class Firework extends Particle
{

	/**
	 * how often a trail is created (percentage from 0 to 1)
	 */
	public static const		TRAIL_PERCENTAGE : Number		= 0.99;

	/**
	 * What color to draw the fire work as
	 */
	public var				color : int					= 0xff9999;

	/**
	 * what alpha
	 */
	public var				alpha : Number				= 1.0;

	/**
	 * A firework is defined by how long till it explodes and by how many rockets are inside of it
	 */
	protected var			seconds_left_until_explosion : Number;

	/**
	 * Payload : how many "sub rockets" burst from the firework when it detonates
	 */
	protected var	 		payload : int;

	/**
	 * Payload max new speed : how "big an explosion". how fast the new rockets will leave the original
	 */
	protected const			payload_max_new_speed : int	= 50;

	/**
	 * The parent GUI is the "drawing" canvas on which the firework "paints" itself.
	 */
	public var				parent_gui : DisplayObjectContainer;

	/**
	 * this is a reference "back" to the simulator that is moving all of the fireworks. it is necessary when we want to
	 * create new fireworks (such as a trail, or when we detonate releasing new fireworks.)
	 */
	protected var 			simulator : Simulator;

	/**
	 * boolean status. Has the firework detonated? This variable is queried by the simulator after all particles have
	 * moved. Any dead fireworks are removed fromt he simulation
	 */
	private var				i_am_dead:Boolean				= false;

	/**
	 * Constructor: Firework
	 * 
	 * Instantiate what it means to be a Firework.
	 * 
	 * Parameters : gui - The parent sprite on which to draw the firework
	 * simulator - The "heart beat" of the program
	 * location - The start X,Y of the firework
	 * velocity - how fast the firework is initially moving
	 * mass - how big the firework is (not really used)
	 * fuse - how long until the firework explodes
	 * payload - how many sub-rockets to explode into
	 * 
	 * Return Value : Constructors don't have return values
	 */
	public function Firework( gui:DisplayObjectContainer, simulator:Simulator, location:Geom_Vector, velocity:Geom_Vector, mass:Number,
						fuse:Number, payload:int )
	{
		// most of what we are is a particle, so use the particles constructor...
		super(location, velocity, mass);

		this.seconds_left_until_explosion = fuse;
		this.parent_gui = gui;
		this.payload = payload;
		this.simulator = simulator;
		this.simulator.add_object_to_simulation(this);

		var gravity : Geom_Vector = Simulator.GRAVITY.clone();
		gravity.scale(mass); // gravity pulls on heavier objects more
		this.add_force(gravity);

	}

	/**
	 * draw the basic "look" of the firework
	 * 
	 * Note: Notice the inverse of the Y value to make the bottom of the screen y==0, instead of the default top
	 * 
	 * Parameters : Graphics - what to draw on
	 * Return Value : None
	 */
	
	public override function paint( graphics:Graphics ) : void
	{
		var inverse_y : int = this.parent_gui.stage.stageHeight - this.location.y;
		graphics.beginFill(color); 
		graphics.drawEllipse(this.location.x, inverse_y, 4,4);
		graphics.endFill();
	}

	/**
	 * When a firework burns up, it is removed from the Simulation. It is up
	 * to the simulator to query this value after a move.
	 * 
	 * Parameters : None
	 * Return Value : None
	 */
	public function burn_up() : void
	{
		this.i_am_dead = true;
	}

	/**
	 * Return if the firework is "Dead" (no longer should be simulated)
	 */
	public override function dead() : Boolean
	{
		return this.i_am_dead;
	}

	/**
	 * Detonate Function:
	 * 
	 * This functions simulates what it means for a firework to explode. A new set of fireworks (the payload) is
	 * released at random velocities from this firework, and then the firework burns up.
	 * 
	 * Parameters : None Return Value : None
	 * 
	 * NOTICE: Velocity Hack
	 * 
	 * Normally when a firework explodes it is at the top of its arc... (has no velocity left)
	 * Usually the audience doesn't see the time it takes to get up there in the sky, but in
	 * our simulation the person is probably wanting the firework to launch from the ground
	 * but get up in the sky and explode quickly... not really, the velocity of the
	 * cannister would carry over to all of the detonated particles, but this doesn't look good
	 * thus:
	 * 
	 * HACK! we change the velocity at explosion to a tenth of what it was, allowing
	 * the firework to Zoom into the sky, but them explode in a nice way
	 */
	protected function detonate() : void
	{
		this.velocity.scale(0.1); // Velocity Hack

		// for every element of the payload, create a new rocket
		for (var i:int = 0; i < this.payload; i++)
		{
			var fuse : Number = (Math.random() + 3.0);
			var firework  : Firework = new Firework(this.parent_gui, this.simulator, this.location, this.velocity, this
					.get_mass(), fuse, 0);
			firework.add_random_velocity();
		}

		this.burn_up();
	}

	/**
	 * This function is a "helper function" to make our payload of new fireworks go in random directions. (?Spherical
	 * random directions)
	 * 
	 * Parameters : None
	 * Return Value : None
	 */
	protected function add_random_velocity() : void
	{
		var radius : Number = Math.random() * this.payload_max_new_speed;
		var angle  : Number = Math.PI * 2 * Math.random();

		alpha = (1.0 - (radius / this.payload_max_new_speed) / 2.0);

		this.color = (alpha * 0xff << 16)+ 0xff0000;

		var add_x : Number = radius * Math.cos(angle);
	    var add_y : Number = radius * Math.sin(angle);

		this.velocity.x += add_x;
		this.velocity.y += add_y;
	}

	/**
	 * What is means to move a firework.
	 * 
	 * This includes two main phases
	 * 
	 * 1) act like a firework: drop a trail. count down (the fuse) until we explode, if appropriate, explode
	 * 2) act like a particle use super.move()
	 * 
	 * Parameters : duration - how long to simulate moving
	 * Return Value : None
	 */
	public override function move( duration:Number ) : void
	{
		seconds_left_until_explosion -= duration;

		if (seconds_left_until_explosion < 0)
		{
			detonate();
			return;
		}
		else
		{
			this.create_trail();
		}

		super.move(duration);

	}

	/**
	 * Create a Trail object... can be overridden by children (such as the Trail object itself, which doesn't produce
	 * trails)
	 * 
	 * Parameters : None
	 * Return Value : None
	 */
	protected function create_trail() : void
	{
		if (Math.random() > TRAIL_PERCENTAGE)
		{
			var fuse : Number = 2 + 3 * Math.random();
			new Trail(this.parent_gui, this, this.simulator, this.location, 10,  fuse);
		}
	}

	/**
	 * Rather than bouncing (the default particle behavior when hitting the ground), we detonate!
	 * 
	 * Parameters : None
	 * Return Value : None
	 */
	protected override function check_for_ground() : void
	{
		if (this.location.y < 0)
		{
			this.detonate();
		}
	}

}

}