package Our_Fireworks;

import java.awt.Color;
import java.awt.Graphics;

/**
 * Author: H. James de St. Germain
 * Date: Fall 2007
 * 
 * This class defines what it means to be a basic particle
 * in a physics simulation. A particle is defined by:
 * 
 * where it is : location
 * where it is going : velocity (includes direciton and speed)
 * its mass : actually the inverse of the mass (for efficiency reasons)
 * 
 * additionally particles have:
 * 
 * forces action on them: at each time step, force can be applied to change
 * the velocity of the particle
 * 
 * when the particle class is combined with the physics simulator, we get
 * a realistic movement
 * 
 * Main functions:
 * 
 * move( duration ) - move for a given time length
 * 
 */

public class Particle implements Movable
{
	/**
	 * location : where in the "real world" we are (As opposed to where on the GUI we are)
	 * velociy : direction AND speed that we are travelling
	 * inverse_mass : 1 / mass (how heavy we are)
	 * force_action_on : the combined forces acting on the particle (usually gravity, etc)
	 * this value modifies the velocity of the particle at each time step
	 */
	protected Geom_Vector	location;
	protected Geom_Vector	velocity;
	private float			inverse_mass;
	private Geom_Vector		force_acting_on;

	/**
	 * The constructor for the particle class.
	 * 
	 * Adds this class to a gui container. sets the start point and velocity, and mass.
	 */
	public Particle( Geom_Vector loc, Geom_Vector vel, float mass )
	{
		this.location = loc.clone();
		this.velocity = vel.clone();
		this.force_acting_on = new Geom_Vector(0, 0);

		this.inverse_mass = 1 / mass;
	}

	/**
	 * At each time step, all applicable forces are applied to the particle
	 * to change its velocity.
	 * 
	 * In this function "something" can "push" or "pull" the particle
	 * by adding a force to it.
	 * 
	 * Note: Forces are additive, so a force in one direction and an equal
	 * force in the opposite direction would cancel each other out.
	 */
	public void add_force( Geom_Vector force )
	{
		force_acting_on.x += force.x;
		force_acting_on.y += force.y;
	}

	/**
	 * We can call this function to remove all forces from this particle.
	 * 
	 */
	public void clear_forces()
	{
		force_acting_on.x = 0;
		force_acting_on.y = 0;
	}

	/**
	 * Here we define the actual PHYSICS of what it means for an object
	 * to be moved and how force affects the velocity of the object.
	 * 
	 * We also check to see if we've hit the ground (in our case Y == 0)
	 * 
	 */
	public void move( float duration )
	{
		if (duration <= 0.0)
			return;

		// update the location based on the velocity (note: not using a^2 / 2)
		this.location.x = this.location.x + (this.velocity.x * duration);
		this.location.y = this.location.y + (this.velocity.y * duration);

		// update velocity based on acceleration
		this.velocity.x = this.velocity.x + (this.force_acting_on.x * this.inverse_mass * duration);
		this.velocity.y = this.velocity.y + (this.force_acting_on.y * this.inverse_mass * duration);

		this.check_for_ground();
	}

	/**
	 * This function defines what it means for a particle to hit
	 * the ground. The basic behavior is for the particle to
	 * "slow down, bounce, and then stop". The "physics" below
	 * are not based on reality, but instead on what "looks good"
	 * on the screen.
	 * 
	 * Also note, the ground is defined as y == 0. This would not
	 * work if there was, say, a mountain to bounce off of.
	 * 
	 * Children classes SHOULD REDEFINE (override) what it means to
	 * hit the ground.
	 */
	protected void check_for_ground()
	{
		/**
		 * bounce off the ground?
		 */
		if (this.location.y < 0)
		{
			this.location.y = 0;
			this.velocity.y = Math.abs(this.velocity.y) / 2.0f;
			this.velocity.x = this.velocity.x / 1.5f;
		}
	}

	/**
	 * Return the mass of the object. Because the mass is stored
	 * as 1/mass (or inverse_mass) we have to take 1/inverse_mass
	 * to get back to the real mass.
	 */
	public float get_mass()
	{
		if (this.inverse_mass == 0)
		{
			return 9999999; // should be infinity
		}

		return 1.0f / this.inverse_mass;
	}

	/**
	 * Modify the location of the particle.
	 */
	public void set_location( float x, float y )
	{
		this.location.x = x;
		this.location.y = y;
	}

	/**
	 * Modify the velocity of the particle.
	 * 
	 * This is usually not done directly, but via moving the particle with
	 * forces applied to it.
	 */
	public void set_velocity( float x, float y )
	{
		this.velocity.x = x;
		this.velocity.y = y;
	}

	/**
	 * get the real world x location of the particle
	 */
	public float get_location_x()
	{
		return this.location.x;
	}

	/**
	 * get the real world y location of the particle
	 */
	public float get_location_y()
	{
		return this.location.y;
	}

	/**
	 * get a copy of the velocity vector of the particle
	 */
	public Geom_Vector get_velocity_vector_copy()
	{
		return this.velocity.clone();
	}

	/**
	 * get the real world x velocity of the particle
	 */
	public float get_velocity_x()
	{
		return this.velocity.x;
	}

	/**
	 * get the real world y velocity of the particle
	 */
	public float get_velocity_y()
	{
		return this.velocity.y;
	}

	/**
	 * draw the basic "look" of a particle
	 * 
	 * Parameters : graphics (what to draw on)
	 * Return Value : None
	 */
	public void paint( Graphics graphics )
	{
		graphics.setColor(Color.black);
		graphics.drawOval((int) this.location.x, (int) this.location.y, 1, 1);
	}

	/**
	 * Create a human readable, concise description of the particle
	 */
	@Override
	public String toString()
	{
		return "Particle@("
				+ this.location.x + ", " + this.location.y + "), " + "vec(" + this.velocity.x + ", " + this.velocity.y
				+ "), " + "mag: " + this.velocity.magnitude() + ", mass: " + this.get_mass();
	}

	/**
	 * use for simulation purposes (if ture, then will be removed from simulation)
	 * 
	 * particles are never dead... inheriting classes can be.
	 */
	@Override
	public boolean dead()
	{
		return false;
	}

}
