package Our_Fireworks;

import java.awt.Color;
import java.awt.Graphics;
import java.util.ArrayList;

import javax.swing.JPanel;


/**
 * Author: H. James de St. Germain
 * Date: Fall 2007
 * 
 * Updated and Ported to Java: Spring 2010
 * 
 * This is a simulator for our physics engine. It has some very
 * basic data and functionality
 * 
 * Data:
 * 1) An array of objects that can move
 * 
 * Functionality:
 * 
 * The simulator allows programs to "Add" particles to the simulation and
 * "remove" particles from it. All particles in the simulation are updated (moved)
 * once per time step.
 * 
 * Physics (animate function):
 * 
 * 1) Add gravity to all particles
 * 2) Move every particle
 * 3) remove forces (prepare for next time step) from each particle
 * 
 * 
 */

public class Simulator
{
	/**
	 * if true, use doubly linked data structure, else
	 * use array list
	 */
	private static boolean				USE_LINKED_LIST	= false;

	/**
	 * CONSTANT: how much gravity to apply
	 */
	public static final Geom_Vector		GRAVITY			= new Geom_Vector(0, -9.8f);

	/**
	 * create and store some information to display the frames per
	 * second that the simulator is able to achieve
	 */
	private long						old_time;

	/**
	 * A poor attempt to compute frames per second
	 */
	private long						frames			= 0;

	/**
	 * remember the real time of when the simulation begins. This is used
	 * to compute how long the simulation takes, frame rate, etc...
	 */
	private long						start_time;

	/**
	 * target frame rate
	 */
	private float						frame_rate_duration;

	/**
	 * keep track of all the objects
	 */
	private ArrayList<Movable>			array;

	/**
	 * the GUI in which the simulation is running
	 */
	private JPanel						containing_sprite;
	
	/**
	 * is the simulation going
	 */
	private boolean						started;

	/**
	 * used for timing large firework computations. This represents the
	 * start time.
	 */
	private long						total_time_start;

	/**
	 * For timing studies, where you are not interested in how much time it takes to draw,
	 * but instead, how much time it takes to do the computations, then set this to
	 * true.
	 */
	private boolean						DO_NOT_DISPLAY	= false;

	/**
	 * Construct the simulator
	 * 
	 * just used to playce the fps display
	 */
	public Simulator( JPanel container_gui, int target_fps )
	{
		this.containing_sprite = container_gui;
		this.frame_rate_duration = 1.0f / target_fps;
		if (USE_LINKED_LIST)
		{
		}
		else
		{
			this.array = new ArrayList<Movable>();
			System.out.println("Array List");
		}

		this.start_time = System.nanoTime();
	}

	/**
	 * Begin running the simulation, one step per frame of the flash movie
	 * that contains the simulator
	 */
	public void start_simulation()
	{
		this.started = true;
	}

	/**
	 * Stop the simulation for now.
	 */
	public void stop_simulation()
	{
		this.started = false;
	}

	/**
	 * add an object to the simulation
	 */
	public void add_object_to_simulation( Movable object )
	{
		if (USE_LINKED_LIST)
		{
		}
		else
		{
			this.array.add(object);
		}
	}

	/**
	 * For every "moveable" object, move it!
	 * 
	 */
	synchronized public void animate()
	{

		if (!this.started)
			return;

		this.frames++;

		if (this.count() > 0)
		{
			if (this.total_time_start == 0)
			{
				this.total_time_start = System.nanoTime();
			}
			System.out.println("Total particles: " + this.count());
		}

		if (this.count() == 0 && total_time_start != 0)
		{
			System.out.println("that took: " + ((System.nanoTime() - total_time_start) / 1000000000.0f) + " seconds.");
			total_time_start = 0;
		}

		// 
		// Physics
		//

		if (USE_LINKED_LIST)
		{
		}
		else
		{
			for (int i = 0; i < this.array.size(); i++)
			{
				Movable current = this.array.get(i);
				current.move(frame_rate_duration); // move for 1/100 of a second simulated time

				if (current.dead())
				{
					this.array.remove(i);
					i--; // go back one spot so we don't skip an item
				}
			}
		}

		this.containing_sprite.repaint();
	}

    /**
     * Draw the simulation.
     * 
     * Every particle is told to draw itself
     * 
     * synchronized means that you cannot be animating and painting at the same time. 
     */ 
	synchronized public void paint( Graphics graphics )
	{
		long new_time = System.nanoTime();

		long elapsed_time = new_time - old_time;
		old_time = new_time;

		long overall_time = new_time - start_time;

		graphics.setColor(Color.white);
		graphics.drawString("Time: " + (1.0f / (elapsed_time / 1000000000.0f)), 20, 20);
		graphics.setColor(Color.white);
		graphics.drawString("FPS: " + (int) (frames / (overall_time / 1000000000.0f)), 20, 40);

		if (DO_NOT_DISPLAY)
		{
			return;
		}

		if (USE_LINKED_LIST)
		{
		}
		else
		{
			for (int i = 0; i < this.array.size(); i++)
			{
				Movable current = this.array.get(i);
				current.paint(graphics);
			}
		}

	}

	/**
	 * @return how many objects in simulator
	 */
	public int count()
	{
		if (USE_LINKED_LIST)
		{
		    return 0;
		}
		else
		{
			return this.array.size();
		}
	}
}
